import os
import json
import asyncio

from typing_extensions import Annotated
import uvicorn
from fastapi import FastAPI, status, HTTPException
from fastapi.responses import StreamingResponse
from pydantic import BaseModel, ConfigDict, Field, BeforeValidator, ValidationError

from .utils import convert_to_bytes, generate_data


class MaxSizePerRequestError(Exception):
    pass


class MinSizePerRequestError(Exception):
    pass


class Testdata:

    class Config(BaseModel):

        model_config = ConfigDict(extra='forbid')

        @staticmethod
        def to_bytes(value: int | str) -> int:
            try:
                return convert_to_bytes(value)
            except Exception as err:
                raise ValidationError from err

        @staticmethod
        def is_authorized_keys(value: set[str] | str) -> set[str]:
            if isinstance(value, str):
                with open(value, encoding='utf-8') as file:
                    return set(filter(lambda x: x.strip() != '', file.read().splitlines()))
            return value

        authorized_keys: Annotated[set[str], BeforeValidator(is_authorized_keys)] = Field(alias='keys')
        max_size: Annotated[int, BeforeValidator(to_bytes)] = Field(alias='max-size')
        max_data: Annotated[int, BeforeValidator(to_bytes)] = Field(alias='max-data')
        buffer_size: Annotated[int, BeforeValidator(to_bytes)] = Field(alias='buffer-size')
        database: str
        update_database_interval: float = Field(alias='update-database-interval', default=5)

    _config: Config
    _api: FastAPI
    _state: dict[str, int]

    def __init__(self, config: Config):
        self._config = config
        self._api = FastAPI(docs_url=None, redoc_url=None)

        # Store internal state
        self._state = {'data-used': 0}

        @self._api.get('/zeros')
        async def zeros(api_key: str, size: int | str) -> StreamingResponse:
            try:
                if api_key not in config.authorized_keys:
                    raise HTTPException(
                        status_code=status.HTTP_401_UNAUTHORIZED,
                        detail='Invalid API Key.'
                    )
                try:
                    size = convert_to_bytes(size)
                except ValueError as err:
                    raise HTTPException(
                        status_code=status.HTTP_400_BAD_REQUEST,
                        detail='Invalid format format for size.'
                    ) from err

                if size < 0:
                    raise MinSizePerRequestError
                if config.max_size < size:
                    raise MaxSizePerRequestError

                # update internal state
                if config.max_data < self._state['data-used'] + size:
                    raise HTTPException(
                        status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                        detail='Service not available.'
                    )
                self._state['data-used'] += size

                return StreamingResponse(
                    status_code=status.HTTP_200_OK,
                    content=generate_data(size, config.buffer_size),
                    media_type='application/octet-stream',
                    headers={
                        'Content-Length': str(size)
                    }
                )

            except MinSizePerRequestError as err:
                raise HTTPException(
                    status_code=status.HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE,
                    detail='Size has to be not-negative.'
                ) from err
            except MaxSizePerRequestError as err:
                raise HTTPException(
                    status_code=status.HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE,
                    detail=f'Exceeded max size per request of {config.max_size} Bytes.'
                ) from err

    async def _update_state(self):
        if os.path.exists(self._config.database):
            mode = 'r+'
        else:
            mode = 'w+'

        with open(self._config.database, mode, encoding='utf-8') as file:
            try:
                self._state = json.load(file)
            except json.JSONDecodeError:
                pass

            while True:
                file.seek(0)
                json.dump(self._state, file)
                file.truncate()
                await asyncio.sleep(self._config.update_database_interval)

    async def run(self, host: str, port: int) -> None:
        try:
            await asyncio.gather(
                asyncio.create_task(uvicorn.Server(uvicorn.Config(self._api, host, port)).serve()),
                self._update_state()
            )
        except asyncio.exceptions.CancelledError:
            pass
